#!/usr/bin/env python

#  Copyright (c) 2004-5 Aaron S. Lav
#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation
#  files (the "Software"), to deal in the Software without
#  restriction, including without limitation the rights to use, copy,
#  modify, merge, publish, distribute, sublicense, and/or sell copies
#  of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:

#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
#ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.



import re
import unicodedata

import codecs

import unihan

class Cedict:
    def __init__ (self, fil):
        ce_regexp = re.compile (r'(.*?) \[(.*)\] /(.*)/')
        self.dict = {}
        self.pinyin_dict = {}
        self.single_chardict = {}
        for l in file (fil).readlines ():
            if l[0] == '#':
                continue
            if l[-2] == '\r': # l[-1] is newline
                l = l[:-2]
            m = ce_regexp.match (l)
            if m == None:
                print "bad cedict line", l
                continue
            big5 = m.group (1)
            pinyin = m.group (2)
            eng = m.group (3).split ('/')
            # 
            def clean (s):
                return s.decode ('ascii', 'replace')
            eng = [clean (trans) for trans in eng]
            try:
                unic = big5.decode ('BIG5')
            except UnicodeError, err:
                print "Unicode decode error", str(err), repr (big5), repr (l)
                continue
            for c in unic:
                # cedict contains a number of ascii punctation marks
                # don't index them.
                if ord (c) < 255:
                    continue
                self.dict.setdefault(c, []).append ((unic, pinyin, eng))
            if len (unic) == 1:
                tup = (unic, pinyin, eng)
                self.single_chardict.setdefault (unic, []).append (tup)
                self.pinyin_dict.setdefault (pinyin, []).append (tup)
                
    def has (self, k):
        return self.get (k) <> None # XXX can optimize by not actually fetching
    
    def get (self, k):
        if ord (k) < 255 or ord (k) > 0xFF00:
            return None
        tracing = False
        rv = self.dict.get (k, None)
        if tracing: print rv
        if rv == None:
            if tracing: print "None", repr (k)
            rv = []
            for d in (unihan.s2t,):
                trans_chars = d.get (k, [])
                if tracing: print "None", repr (k), repr(trans_chars)
                for ch in trans_chars:
                    defs = self.dict.get (ch, [])
                    rv = rv + defs
                if tracing: print "None", len (rv)
            if len (rv) == 0:
                return None
            return rv
        return rv


    def get_singlechar (self, k):
        rv = self.single_chardict.get (k, None)
        if rv <> None:
            return rv
        for d in unihan.s2t, unihan.t2s:
            trans_chars = d.get (k, [])
            for ch in trans_chars:
                rv = self.single_chardict.get (ch, None)
                if rv <> None:
                    # early-out: callers should perhaps be modified to handle all chars
                    return rv
        return None
            
        
    def dump (self):
        it = self.pinyin_dict.items ()
        it.sort ()
        for k, v in it:
            print k
            v.sort ()
            for (u, p, e) in v:
                print "   ", p, e
        for (k, v) in self.dict.iteritems ():
            print repr (k)
            for t in v:
                (unic, pinyin, eng) = t
                print  "\t", repr (unic), pinyin, eng

num_to_combining_dict = {
    1: u"\u0304", # macron
    2: u"\u0301", # acute
    3: u"\u030c", # caron / hacek
    4: u"\u0300", # grave
    5: ""}

# http://www.romanization.com/pinyintonemarks/
# A and e trump all other vowels and always take the tone mark. There are no Mandarin syllables that contain both a and e.
# In the combination ou, o takes the mark.
# In all other cases, the final vowel takes the mark.

def pinyin_num_to_tones (pinyin_num):
    tone_num = pinyin_num[-1]
    if not tone_num.isdigit ():
        return pinyin_num
    pinyin = unicodedata.normalize ('NFD', unicode (pinyin_num[:-1]))
    # decompose to ease handling of u-umlaut
    
    split_list = ['a', 'e', 'ou']
    index = -1
    for s in split_list:
        index = pinyin.find (s)
        if index <> -1:
            break
    vowel_list = ['i', 'o', 'u']
    if index == -1:
        for vowel in vowel_list:
            index = pinyin.rfind (vowel)
            if index <> -1:
                break
    if index == -1:
        print "vowel not found for", repr (pinyin)
        return pinyin_num
    start = pinyin [:index + 1]
    end = pinyin [index + 1:]
    tone = num_to_combining_dict.get (int(tone_num), '')
    return unicodedata.normalize ('NFC', start + tone + end)
            
if __name__ == '__main__':
    cedict = Cedict ('cedict.b5')
    cedict.dump ()
