#!/usr/bin/env python

#  Copyright (c) 2004-5 Aaron S. Lav
#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation
#  files (the "Software"), to deal in the Software without
#  restriction, including without limitation the rights to use, copy,
#  modify, merge, publish, distribute, sublicense, and/or sell copies
#  of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:

#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
#ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.




class MediaType:
    """Implements Internet Media Types parsing (see RFC1521 7.4.3., RFC1590,
    and RFC2616 3.7)"""
    def __init__ (self, val, url):
        a = val.split (';')
        typ_arr = a[0].split ('/')
        try:
            self.typ = (typ_arr[0].strip (), typ_arr[1].strip())
        except IndexError:
            if a[0] == 'text-html':
                self.typ = ('text', 'html')
                # see http://www.bestcatering.com.sg, whose 404 error
                # documents contain meta http-equiv content-type
                # "text-html; charset=Windows-1252"
                # Note that this site generates a utf-8 encoding
                # BOM before the META GB2312 declaration is encountered,
                # while the headers have no charset declaration and
                # thus we should be parsing iso8859-1.  The food looks good,
                # though.
            else:
                print "bad MediaType 2", val, url
                self.typ = ('unknown', 'unknown')
        self.parms = {}
        for kv in a[1:]:
            t = kv.split ('=')
            try:
                self.parms [t[0].strip ()] = t[1].strip ()
            except IndexError:
                print "Mediatype err", kv
    def to_inet (self):
        s_l = ["%s/%s" % (self.typ[0], self.typ[1])]
        if len (self.parms) <> 0:
            s_l.append (';')
            for k, v in self.parms.items ():
                s_l.append ('%s=%s' % (k, v))
        return "".join (s_l)
        
    def __str__ (self):
        return "%s/%s p:%s" % (self.typ[0], self.typ[1], str (self.parms))
    
