#  Copyright (c) 2004-5 Aaron S. Lav
#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation
#  files (the "Software"), to deal in the Software without
#  restriction, including without limitation the rights to use, copy,
#  modify, merge, publish, distribute, sublicense, and/or sell copies
#  of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:

#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
#ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.


simplevar = 'kSimplifiedVariant'
traditionalvar = 'kTraditionalVariant'

simple_to_trad = {}
trad_to_simple = {}
mandarin = {}
defn = {}
compatibility = {}

# 'kRSUnicode'
char_to_rad = {}
rad_to_char = {}

# 'kPhonetic'
char_to_phonetic = {}
phonetic_to_char = {}


# If a character is both a traditional variant and a simplified variant
# of another character, it's listed as its own simplified variant
# and one of its own traditional variants

def unichr_surrogate (cp):
    """Encode a code point using the UTF-16 surrogate algorithm if
    necessary.  See <http://www.unicode.org/faq/utf_bom.html#34>, from
    which some of the code below is lifted.

    This is both unnecessary and technically incorrect if we're running a
    wide build (UCS-4) of Python: I should figure out if there's a
    better way to tell than seeing if unichr(0x10000) throws an exn."""
    if cp <= 0xFFFF:
        return unichr (cp)
    LEAD_OFFSET = 0xD800 - (0x10000 >> 10)
    lead = LEAD_OFFSET + (cp >> 10);
    trail = 0xDC00 + (cp & 0x3FF);
    return unichr (lead) + unichr (trail)
                             
def parse (fn):
    for line in open (fn):
        if line[0] == '#':
            continue
        if len (line) == 0:
            continue
        if line[-1] == '\n':
            line = line[:-1]
        char_spec, typ, info = line.split ('\t', 3)
        if char_spec [0:2] <> 'U+':
            print "bad line", line
            continue
        def decode_charspec (s):
            val = int(s[2:], 16)
            return unichr_surrogate (val)
        
        char_spec = decode_charspec (char_spec)

        if typ == 'kMandarin':
            mandarin [char_spec] = info
        elif typ == 'kDefinition':
            defn [char_spec] = info.decode ('utf-8')
        elif typ == 'kCompatibilityVariant':
            compatibility[char_spec] = decode_charspec (info)
        elif typ == 'kRSUnicode':
            def trim_radical (s):
                r = s.split ('.')[0] # ignore stroke-count
                if r[-1] == "'": # ignore simplified flag
                    r = r[:-1]
                return r
            radicals = info.split (' ')
            radicals = [trim_radical (radical) for radical in radicals
                        if radical <> '']
            for radical in radicals:
                char_to_rad.setdefault (char_spec, []).append (radical)
                rad_to_char.setdefault (radical, []).append (char_spec)
        elif typ == 'kPhonetic':
            phonetics = info.split (' ')
            for phonetic in phonetics:
                if phonetic == '':
                    continue
                char_to_phonetic.setdefault (char_spec, []).append (phonetic)
                phonetic_to_char.setdefault (phonetic, []).append (char_spec)

        elif typ == simplevar:
            # 'Unicode value for the simplified Chinese variant for this char'
            if trad_to_simple.has_key (char_spec):
                print "already has", trad_to_simple[char_spec], "now", info
            trad_to_simple[char_spec] = map (decode_charspec,info.split (' '))
        elif typ == traditionalvar:
            if simple_to_trad.has_key (char_spec):
                print "already has", simple_to_trad[char_spec], "now", info
            simple_to_trad[char_spec] = map (decode_charspec,info.split (' '))

def count_mult (d):
    count = 0
    k_list = []
    for k, v in d.iteritems ():
        if len (v) > 1:
            count += 1
            k_list.append (k)
    return k_list, count


def write_multi_xlation (d, title, fn):
    d_list, _ = count_mult (d)
    ofil = open (fn, 'w')
    ofil.write ("""
<html><head><meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
<title>%s</title></head><body><p><ul>""" % (title,))

    def write_list (l):
        for k in l:
            simple_list = [simple.encode ('utf-8') for simple in d [k]]
        
            ofil.write ("""<li>%s %s %s %s</li>""" % (
                k.encode ('utf-8'), mandarin [k], defn[k].encode ('utf-8'),
                " ".join (simple_list)))
        ofil.write ('</ul>')

    write_list (d_list)
    d2 = [v for v in d_list if v in d[v]]
    write_list (d2)
    ofil.write ('</body></html>')
    ofil.close ()

def write_py (fil, nm, d):
    fil.write ('%s = {\n'  % nm)
    l = ["%s: %s" % (repr (k), repr (v)) for k, v in d.iteritems ()]
    fil.write (',\n'.join (l))
    fil.write ('\n}\n')
    


if __name__ == '__main__':
    parse ('/tmp/Unihan.txt')

    print "t2s", len (trad_to_simple)
    print "s2t", len (simple_to_trad)
    print "compat", len (compatibility)

    write_multi_xlation (trad_to_simple,
                         'Traditional To Simplified', 
                         '../t2s.html')
    
    write_multi_xlation (simple_to_trad,
                         'Simplified To Traditional', 
                         '../s2t.html')

    fil = open ('unihan.py', 'w')
    write_py (fil, 's2t', simple_to_trad)
    write_py (fil, 't2s', trad_to_simple)
    write_py (fil, 'compat', compatibility)
    write_py (fil, 'c2r', char_to_rad)
    write_py (fil, 'r2c', rad_to_char)
    write_py (fil, 'c2p', char_to_phonetic)
    write_py (fil, 'p2c', phonetic_to_char)

    fil.close ()

