#  Copyright (c) 2004-5 Aaron S. Lav
#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation
#  files (the "Software"), to deal in the Software without
#  restriction, including without limitation the rights to use, copy,
#  modify, merge, publish, distribute, sublicense, and/or sell copies
#  of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:

#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
#ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.


import urlparse
import sys
import traceback
import cgi

from twisted.protocols import http

import rewriter
import globalvars
import stats
import unihan
import cedict


class MyRequest(http.Request):
    
    def error (self, text):
        self.write ('<html><head><title>Error</title></head><body><p>\n')
        self.write (text)
        self.write ('</p></body></html>')


    def process_stats (self, pathl):
        self.write ('<html><head><title>Stats</title></head><body><p>\n')
        self.write ('<ul>\n')
        if pathl[0] == '':
            pathl = pathl[1:]
        if len (pathl) == 0:
            s_l = stats.stats.items ()
            s_l.sort ()
            for k, v in s_l:
                self.write ('<li><a href="%s">%s</a> %d</li>' % (
                    k, k, len (v)))
        else:
            d = stats.stats[pathl[0]]
            it = d.items ()
            it = [(v, k) for k, v in it]
            it.sort ()
            it.reverse ()
            it = [(k, v) for v, k in it]
            for k, v in it:
                self.write ('<li>%s %d</li>' % (k, v))
        self.write ('</ul></p></body></html>')

    def process_chars (self, char_id, prev_id, next_id):
        try:
            char = unichr (int (char_id, 16))
            prev = unichr (int (prev_id, 16))
            next = unichr (int (next_id, 16))
        except ValueError, err:
            self.error ('Bad char_id: %s %s %s' % (char_id, prev, next))
            return
            
        char_info = globalvars.chinese_dict.get (char)
        if char_info == None:
            self.error ('No info for char %s %s' % (char_id,
                                                    char.encode ('utf-8')))
            return

        stats.inc_stats (stats.STAT_LOOKUPCHARS, char.encode ('utf-8'))
        self.write ('<html><head><title>%s</title></head><body><p><ul>' % (
            char.encode('utf-8')))
            
        def key (c):
            uni, pinyin, eng = c
            match = 0
            ind = uni.find (char)
            if ind == -1:
                trans_l = unihan.s2t.get (char, [])
                for trans_c in trans_l:
                    ind = uni.find (trans_c)
                    if ind <> -1:
                        break
                    # if can't find at all, sort after everything else
                    return (1, len (uni), 'zzz')
            if uni[ind-1:ind] == prev:
                match += 1
            if uni[ind+1:ind+2] == next:
                match += 1
            # we sort syllable-by-syllable, first
            # by pinyin syllable, and, within pinyin syllable,
            # break ties by character
            pinyin_l = pinyin.split (' ')
            orderer = zip (pinyin_l, uni)
            # if the dictionary is erroneous and there's a mismatch
            # between the number of chars and the number of pinyin
            # syllables, not clear what the right thing is
                
            return (-1 * match, len (uni), orderer)

        char_info_tosort = [(key(c), c) for c in char_info]
        char_info_tosort.sort ()
        
        for (_, char_info) in char_info_tosort:
            (uni, pinyin, eng) = char_info
            l2 = []
            prev = unichr (0) + uni
            next = uni[1:] + unichr (0)
            for c, p, n in zip (uni, prev, next):
                s = rewriter.make_link (c, p, n, globalvars.chars_prefix)
                l2.append (s)

            pinyin_urls = self.make_pinyin_urls (pinyin)
            eng_str = (u" / ".join (eng)).encode ('utf-8')
            self.write ('<li>%s %s %s</li>\n' % (
                "".join (l2).encode ('utf-8'), pinyin_urls, eng_str))
        self.write ('</ul></p>')
        self.write ('<p>')
        def rad_phone (d, prefix, descr):
            rp_list = d.get (char, [])
            def mk_text (rp):
                return '<a href="%s/%s">%s %s</a>' % (
                    prefix, rp, descr, rp)
            text = "<br />\n".join (map (mk_text, rp_list))
            self.write (text.encode ('utf-8'))
        rad_phone (unihan.c2r, globalvars.radical_prefix, 'Radical')
        self.write ('<br />')
        rad_phone (unihan.c2p, globalvars.phonetic_prefix, 'Phonetic')
        self.write ('</p>')
        self.write ('<p>Chinese-English dictionary provided by <a href="http://www.mandarintools.com/cedict.html">CEDICT</a></p>')
        self.write ('<p>Web interface by <a href="http://www.pobox.com/~asl2/software/middleproxy">Middleproxy</a></p>')
        self.write ('<p><a href="/stats/">Statistics</a></p>')
        self.write ('</body><html>')


    def make_pinyin_urls (self, pinyin):
        l = pinyin.split (' ')
        def one_url (syllable):
            if syllable == '':
                return ''
            return '<a href="%s/%s">%s</a>' % (
                globalvars.pinyin_prefix,
                syllable,
                cedict.pinyin_num_to_tones(syllable).encode ('utf-8'))
        return " ".join ([one_url(syl) for syl in l])
    def process_pinyin (self, pathl):
        pinyin = pathl [0]
        self.write ('<html><head><title>%s</title></head><body><p><ul>' % (
            pinyin))
        char_info = globalvars.chinese_dict.pinyin_dict.get (pinyin)
        for ci in char_info:
            u, _, eng = ci
            c = u [0]
            self.write ('<li>')
            engs = "/ ".join (eng)
            s = u'%s %s' % (
                rewriter.make_link (c, chr(0), chr(0),
                                    globalvars.chars_prefix),
                 engs)
            self.write (s.encode ('utf-8'))
            self.write ('<br>')
            self.write ('<br>')
            self.write ('</li>')
            
        self.write ('</body></html>')

    def process_radphon (self, pathl, rp):
        index = pathl [0]
        self.write ('<html><head><title>%s</title></head><body><p><ul>' % (
            index))
        chars = rp.get (index, [])
        for char in chars:
            if len (char) > 1:
                # XXX a narrow build of Python and a char outside the BMP
#                print "bad char", repr (char)
                continue
            self.write ('<li>')

            s = rewriter.make_link (char, chr(0), chr (0), globalvars.chars_prefix)
            self.write (s.encode ('utf-8'))
            singlechar_info = globalvars.chinese_dict.get_singlechar (char)
            if singlechar_info <> None:
                def one_defn (info):
                    uni, pinyin, eng = info
                    tone_pinyin = cedict.pinyin_num_to_tones (pinyin)
                    link_pinyin = '<a href="%s/%s">%s</a>' % (
                        globalvars.pinyin_prefix, pinyin, tone_pinyin)
                    return " " + link_pinyin + " " + "/".join (eng)
                text = "; ".join (
                    [one_defn (info) for info in singlechar_info])
                self.write (text.encode ('utf-8'))


            self.write ('</li>')
            
        self.write ('</body></html>')
        
    def process (self):
        (scheme, netloc, path, parms, query, frag
         ) = urlparse.urlparse (self.uri)

        self.setHeader ('Content-Type', 'text/html; charset=utf-8')

        query_dict = http.parse_qs (query)
        pathl = path.split ('/')
        def get (s):
                # take 0th occurrence of parm in string, in each case
                return query_dict.get (s, ['0'])[0]
        try:
            if pathl[0] == '':
                pathl = pathl[1:]
            if path == '/chars':
                self.process_chars (get('char'), get('prev'), get ('next'))
            elif pathl[0] == 'stats':
                self.process_stats (pathl[1:])
            elif pathl[0] == 'pinyin':
                self.process_pinyin (pathl[1:])
            elif pathl[0] == 'radical':
                self.process_radphon (pathl[1:], unihan.r2c)
            elif pathl[0] == 'phonetic':
                self.process_radphon (pathl[1:], unihan.p2c)
            else:
                self.write ('<html><head><title>Error</title></head><body>Unimplemented URL %s</body></html>' % (self.uri,))
        except:
            typ, val, tb = sys.exc_info ()
            tb_list = map (cgi.escape, traceback.format_tb (tb))
            tbstr = "<br />\n".join (tb_list)
            
            self.write ('<html><head><title>Exception</title></head><body>')
            self.write ('<p>Error processing %s</p>' % cgi.escape ((self.uri)))
            self.write ('<p>Exception: %s %s</p>' % (
                cgi.escape (str (typ)), cgi.escape (str (val))))
            self.write ('<p>Traceback: %s</p>' % tbstr)
            self.write ('''<p>Please contact asl2@pobox.com for problems or see <a href="http://www.pobox.com/~asl2/software/middleproxy">Middleproxy's</a> home page</p>''')
            self.write ('</body></html>')
        self.finish ()

class MyHttpProtocol(http.HTTPChannel):
    requestFactory = MyRequest
