#!/usr/bin/env python

#  Copyright (c) 2004-5 Aaron S. Lav
#  Permission is hereby granted, free of charge, to any person
#  obtaining a copy of this software and associated documentation
#  files (the "Software"), to deal in the Software without
#  restriction, including without limitation the rights to use, copy,
#  modify, merge, publish, distribute, sublicense, and/or sell copies
#  of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:

#  The above copyright notice and this permission notice shall be
#  included in all copies or substantial portions of the Software.

#THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
#EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
#MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
#NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
#BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
#ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
#CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#SOFTWARE.



import re


import scraper
import globalvars
import unihan
import cedict

def hexif (c):
    return hex (ord(c))[2:]

do_js = 1
debug_missing = 0

def fold_char (c):
    c = unihan.compat.get (c, c)
    return c

def make_link (c, prev, next, base):
    js = ""
    if do_js:
        singlechar_info = globalvars.chinese_dict.get_singlechar (c)
        if singlechar_info <> None:
            def one_defn (info):
                uni, pinyin, eng = info
                pinyin = cedict.pinyin_num_to_tones (pinyin)
                return pinyin + " " + "/".join (eng)
            text = "; ".join ([one_defn (info) for info in singlechar_info])
            js = '''onMouseOver="window.status='%s'; return true"
            onMouseOut="window.status=''; return true"''' % (text)


    return '<a href="%s%s&prev=%s&next=%s" %s>%s</a>' % (
        base, 
        hexif(c),
        hexif(prev), hexif(next),
        js,
        c)

class RewritingScraper(scraper.Scraper):
    def __init__ (self, linkstr):
        self.linkstr = linkstr
        scraper.Scraper.__init__ (self)
        self.rewriting = 0
        # the "rewriting" flag is used to turn off rewriting
        # until we get to the body, because most browsers
        # don't take kindly to <a> tags in the title.
        self.in_a = 0
        # are we currently rewriting an <a> tag?
    def pdata (self, inchunk):
        if not self.rewriting:
            return inchunk

        inchunk_l = map (fold_char, inchunk)
        prev_l = [chr(0)] + inchunk_l [:-1]
        next_l = inchunk_l [1:] + [chr(0)]
        
        def proc_onechar (c, prev, next):
            if globalvars.chinese_dict.has (c):
                return make_link (c, prev, next, globalvars.mycharurl)
            if debug_missing and ord (c) > 0x100:
                return '<abbr title="%s">%s</abbr>' % (hex (ord (c)), c)
            return c
        rv =  "".join ([proc_onechar (c, p, n) for c, p, n in
                        zip (inchunk_l, prev_l, next_l)])
        return rv
    def endtag (self, thetag):
        if thetag.lower () <> '/a' or not self.in_a:
            return scraper.Scraper.endtag (self, thetag)
        self.in_a = 0
        return ""
    def handletag(self, name, attrs, thetag):
        if name == 'body':
            self.rewriting = 1
        rv = scraper.Scraper.handletag (self, name, attrs, thetag)
        # will browsers render a href w/in a name (name is fragment tag?)
        # note: will this work on <a href="" ... > ... <a name="adfadsf">.</a>.</a>?
        # note: will fail on <a href="..." ...> <a href="..." ...> </a>,
        # i.e. missing end </a> tag.
        if name.lower () == 'a' and not self.in_a:
            for key, val in attrs:
                if key.lower () == 'href':
                    self.in_a = 1
                    return rv + self.linkstr + "</a>"
        return rv

escape_re = re.compile ("(?P<dec>&#\d*;)|(?P<hex>&#x[\da-fA-F]*;)")

def unescape (s):
    def repl (mo):
        d = mo.groupdict ()
        val = d['dec']
        base = 10
        offset = 2
        if val == None:
           val = d['hex']
           base = 16
           offset = 3

        s = val[offset:-1] # remove trailing ';'
        rv = int (s, base)
        return unichr (rv)
    
    return escape_re.sub (repl, s)
    

def rewrite (inp, link_str):
    """Takes unicode input, and rewrites chinese char to
    <a href="(def link)">char</a>, and <a ...> [foo]</a> to
    <a ... >LINK</a>[foo].  Interprets &...; escapesReturns Unicode"""

    inp = unescape (inp)

    Scraper = RewritingScraper (link_str)
    Scraper.feed (inp)
    return Scraper.close ()
    

    
if __name__ == '__main__':
    import sys
    fin = open (sys.argv[1], 'rb')
    fout = open (sys.argv[2], 'wb')
    inp = fin.read().decode ('gb2312')
    out = rewrite (inp, 'LINK')
    fout.write (out.encode ('utf-8'))
    

